/*
 * Created on Tue Mar 07 2023
 *
 * Copyright (C) 2023 PixelApps - All Rights Reserved
 * The code in this file is Intellectutal Property of PixelApps
 * and can't be edited, redistributed or used without a valid license of it's product (Welltested)[https://welltested.ai]
 */
import 'package:analyzer/dart/ast/ast.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:build/build.dart';
import 'package:source_gen/source_gen.dart';
import 'package:welltested/src/executors/fixer_executor.dart';
import 'package:welltested/src/executors/path_executor.dart';
import 'package:welltested/src/models/test_request_data.dart';
import 'package:welltested/src/repositories/dart_server_repository.dart';
import 'package:welltested/src/repositories/generations_repository.dart';
import 'package:welltested/src/repositories/writer_repository.dart';
import 'package:welltested/src/utilities/utilities.dart';
import 'package:welltested/src/visitors/class_visitor.dart';
import 'package:welltested/src/visitors/function_body_visitor.dart';

Future<String> generateWelltestedSource(
  BuildStep buildStep,
  ClassElement element, {
  required FixerExecutor fixerExecutor,
  required DartServerExecutor dartServerExecutor,
  ConstantReader? annotation, // @Samyak Can we make it `required`?
  bool save = false,
  required PathExecutor pathExecutor,
}) async {
  final List<String> excludedMethods = annotation
          ?.read('excludedMethods')
          .listValue
          .map((e) => e.toStringValue() ?? "")
          .toList() ??
      [];

  final fileAstNode = (await buildStep.resolver.astNodeFor(element))?.parent;
  final fileImport =
      LibraryReader(pathExecutor.inputLibrary).pathToElement(element);

  final GenerationRepository testExecutor = GenerationRepository();
  final WriterRepository writerExecutor = WriterRepository(buildStep);
  final classVisitor = ClassVisitor();
  element.visitChildren(classVisitor);

  final Map<String, MethodDeclaration> methodAstNodes = {};
  for (final method in classVisitor.methods.entries) {
    final node = (await getResolvedAstNodeForElement(method.value, buildStep))
        as MethodDeclaration;
    methodAstNodes[method.key] = node;
  }
  try {
    for (int i = 0; i < methodAstNodes.entries.length; i++) {
      final method = methodAstNodes.entries.elementAt(i);

      if (excludedMethods.contains(method.key))
        continue; //Don't generate for excluded methods

      final ast_method = method.value;
      final FunctionBodyVisitor functionBodyVisitor =
          FunctionBodyVisitor(pathExecutor: pathExecutor);
      ast_method.visitChildren(functionBodyVisitor);

      final List<ClassDeclaration> modelNodes = [];
      for (final model in functionBodyVisitor.models) {
        final node =
            await buildStep.resolver.astNodeFor(model) as ClassDeclaration;
        modelNodes.add(node);
      }

      final TestRequestData testRequestData = TestRequestData(
        fileNode: fileAstNode!,
        functionName: method.key,
        methodAstNodes: methodAstNodes,
        classType: classVisitor.className,
        fileImportPath: fileImport,
        modelNodes: modelNodes,
        callFlows: functionBodyVisitor.callFlows,
        variableDeclarations: functionBodyVisitor.updateVariableNameNType,
      );

      if (save) {
        final file =
            await pathExecutor.correspondingPermanentTestFile(method.key);
        if (file == null) {
          continue;
        }

        await testExecutor.updateFixedTest(
            testRequestData, await file.readAsString());
      } else {
        log.info(" Generating Test for ${testRequestData.functionName}");
        final unitTests = await testExecutor.buildUnitTests(testRequestData);

        final overrideResult = unitTests['fixed_completion'];
        if (overrideResult != null) {
          final file =
              await pathExecutor.correspondingPermanentTestFile(method.key);
          if (file != null && (file.readAsString() == overrideResult)) {
            continue;
          }
          final asset = AssetId(pathExecutor.inputLibraryInfo.package,
              pathExecutor.correspondingTemporaryTestFilePath(i));

          await writerExecutor.outputGeneratedFile(
              asset, overrideResult, method.key);
          continue;
        }
        if (unitTests['model_completion'] == null) {
          log.warning(
              "⛔️ Sorry, we can't process ${testRequestData.className} at the moment. Please try again Later.");
          continue;
        }

        final file = await writerExecutor.writeTestToCorrespondingFile(
            testRequestData.fileImportPath,
            unitTests['model_completion'],
            method.key);
        await dartServerExecutor.addFilesToRoot([file.absolute.path]);
        final updatedContent =
            await fixerExecutor.findAndFixIssues(file.path, method.key);
        final asset = AssetId(pathExecutor.inputLibraryInfo.package,
            pathExecutor.correspondingTemporaryTestFilePath(i));
        await writerExecutor.outputGeneratedFile(
            asset, updatedContent, method.key);
        await file.delete();
        log.info(" Test Generated for ${testRequestData.functionName}");
      }
    }
  } catch (e) {
    throw (e);
  } finally {
    dartServerExecutor.dispose();
  }
  return '';
}
